// MockRobotiq2fSocketAdapter.hpp

#ifndef MOCK_ROBOTIQ2F_SOCKET_ADAPTER_HPP
#define MOCK_ROBOTIQ2F_SOCKET_ADAPTER_HPP

#include <string>
#include <map>
#include <atomic>
#include <tuple>

#include "robotiq_2f_interface/SocketAdapterBase.hpp"

namespace robotiq_interface
{
class MockRobotiq2fSocketAdapter : public SocketAdapterBase {
public:
        MockRobotiq2fSocketAdapter() : isConnected(false), gripperPosition(0), gripperForce(0), gripperSpeed(0), isActive(false) {}

    ~MockRobotiq2fSocketAdapter() override {
        if (movementThread && movementThread->joinable()) {
            movementThread->join();
        }
    }

    bool connecting(const std::string& hostname, int port) override {
        isConnected.store(true);
        return true; // Simulate successful connection
    }

    void disconnect() override {
        isConnected.store(false);
    }

    int force() override {
        return gripperForce.load();
    }

    int speed() override {
        return gripperSpeed.load();
    }

    int position() override {
        return gripperPosition.load();
    }

    bool is_active() override {
        return isActive.load();
    }

    void activate(bool autoCalibrate = true) override {
        isActive.store(true);  // Ensure we set active before moving
        if (!autoCalibrate) {
            // Simulate a calibration sequence
            move_and_wait(0, 255, 1);
        }
    }

    void deactivate() override {
        isActive.store(false);
    }

    std::tuple<bool, int> move(int targetPosition, int speed, int force) {
        // Start a new movement operation in a background thread
        if (movementThread && movementThread->joinable()) {
            movementThread->join(); // Ensure no concurrent movements
        }

        gripperSpeed.store(speed);  // Simulate immediate speed update
        gripperForce.store(force);  // Simulate immediate force update

        movementThread = std::make_unique<std::thread>(&MockRobotiq2fSocketAdapter::simulateMovement, this, targetPosition, speed, force);
        return std::make_tuple(true, targetPosition);
    }

    void move_and_wait(int targetPosition, int speed, int force) {
        auto [success, pos] = move(targetPosition, speed, force);
        if (movementThread && movementThread->joinable()) {
            movementThread->join();  // Wait for movement to complete
        }
    }

private:
    std::atomic<bool> isConnected;
    std::atomic<int> gripperPosition;
    std::atomic<int> gripperForce;
    std::atomic<int> gripperSpeed;
    std::atomic<bool> isActive;

    std::unique_ptr<std::thread> movementThread;

    void simulateMovement(int targetPosition, int speed, int force) {
        int step = (targetPosition > gripperPosition.load()) ? 1 : -1;

        // Simulate moving towards the target position
        while (gripperPosition.load() != targetPosition) {
            gripperPosition.fetch_add(step);
            std::this_thread::sleep_for(std::chrono::milliseconds(10)); // Simulate time to move a step
        }

        // std::cout << "Reached position " << targetPosition << " with speed " << speed << " and force " << force << std::endl;
    }
};
} // end robotiq_interface

#endif // MOCK_ROBOTIQ2F_SOCKET_ADAPTER_HPP