// hardware_interface.hpp

#ifndef ROBOTIQ2F_HARDWARE_INTERFACE_HPP
#define ROBOTIQ2F_HARDWARE_INTERFACE_HPP

#include <chrono>
#include <cmath>
#include <limits>
#include <memory>
#include <vector>
#include <atomic>

#include "robotiq_2f_interface/visibility_control.hpp"

#include "robotiq_2f_interface/SocketAdapterBase.hpp"
#include "robotiq_2f_interface/Robotiq2fSocketAdapter.hpp"
#include "robotiq_2f_interface/MockRobotiq2fSocketAdapter.hpp"

#include <hardware_interface/handle.hpp>
#include <hardware_interface/hardware_info.hpp>
#include <hardware_interface/system_interface.hpp>
#include <hardware_interface/types/hardware_interface_type_values.hpp>
#include <hardware_interface/types/hardware_interface_return_values.hpp>

#include <rclcpp/rclcpp.hpp>
#include <rclcpp/clock.hpp>

#include <stdexcept>

namespace robotiq_interface
{
class RobotiqGripperHardwareInterface : public hardware_interface::SystemInterface
{
public:
    // Constructors
    ROBOTIQ_INTERFACE_PUBLIC RobotiqGripperHardwareInterface();
    ROBOTIQ_INTERFACE_PUBLIC ~RobotiqGripperHardwareInterface();
    // ROBOTIQ_DRIVER_PUBLIC explicit RobotiqGripperHardwareInterface(std::unique_ptr<SocketFactory> socket_factory);

    // Lifecycle Management
    ROBOTIQ_INTERFACE_PUBLIC CallbackReturn on_init(const hardware_interface::HardwareInfo& info) override;
    ROBOTIQ_INTERFACE_PUBLIC CallbackReturn on_configure(const rclcpp_lifecycle::State& previous_state) override;
    ROBOTIQ_INTERFACE_PUBLIC CallbackReturn on_activate(const rclcpp_lifecycle::State& previous_state) override;
    ROBOTIQ_INTERFACE_PUBLIC CallbackReturn on_deactivate(const rclcpp_lifecycle::State& previous_state) override;

    // State and Command Interfaces
    ROBOTIQ_INTERFACE_PUBLIC std::vector<hardware_interface::StateInterface> export_state_interfaces() override;
    ROBOTIQ_INTERFACE_PUBLIC std::vector<hardware_interface::CommandInterface> export_command_interfaces() override;

    // Data Access (Read/Write)
    ROBOTIQ_INTERFACE_PUBLIC hardware_interface::return_type read(const rclcpp::Time& time, const rclcpp::Duration& period) override;
    ROBOTIQ_INTERFACE_PUBLIC hardware_interface::return_type write(const rclcpp::Time& time, const rclcpp::Duration& period) override;

protected: // Likely changes the access to protected from private
    // Members for driver interaction
    bool use_mock_hardware_;
    std::unique_ptr<SocketAdapterBase> socket_adapter_;
    void configureAdapter(bool useMock);

    // Background communication thread
    std::thread communication_thread_;
    std::atomic<bool> communication_thread_is_running_;
    std::mutex resource_mutex_;
    void background_task();
    void performRegularOperations();

    static constexpr double NO_NEW_CMD_ = std::numeric_limits<double>::quiet_NaN();

    // Conversion Methods
    double convertRawToGap(int raw_position);
    double convertRawToAngle(int raw_position);
    unsigned int convertGapToRaw(double gap);
    unsigned int convertSpeedToRaw(double speed);
    unsigned int convertForceToRaw(double force);

    // Gripper Commands
    std::atomic<uint8_t> write_command_;
    std::atomic<uint8_t> write_force_;
    std::atomic<uint8_t> write_speed_;
    std::atomic<uint8_t> gripper_current_state_;


    // Gripper State Tracking
    double gripper_gap_ = 0.0;
    double gripper_angle_ = 0.0;
    double gripper_velocity_ = 0.0;
    double gripper_gap_command_ = 0.0;

    rclcpp::Clock::SharedPtr clock_ = std::make_shared<rclcpp::Clock>(RCL_ROS_TIME);
    rclcpp::Time last_update_time_ = clock_ -> now();
    rclcpp::Time current_time_ = clock_ -> now();

    double reactivate_gripper_cmd_ = 0.0;
    std::atomic<bool> reactivate_gripper_async_cmd_;
    double reactivate_gripper_response_ = 0.0;
    double gripper_force_ = 0.0;
    double gripper_speed_ = 0.0;
    std::atomic<std::optional<bool>> reactivate_gripper_async_response_;

    // Connection Variables
    std::string robot_ip_;
    int robot_port_;

    // Parameters for physical limits and configuration
    double min_position_;
    double max_position_;
    double max_angle_;
    double min_speed_;
    double max_speed_;
    double min_force_;
    double max_force_;

    // Last Command Values
    int last_cmd_gap = 0;
    int last_cmd_speed = 0;
    int last_cmd_force = 0;
    int last_raw_gap = 0;

    // loop time
    const std::chrono::milliseconds gripperCommsLoopPeriod = std::chrono::milliseconds{ 2 };

    // Logger
    const rclcpp::Logger logger_ = rclcpp::get_logger("RobotiqGripperHardwareInterface");
};
} // end robotiq_interface

#endif //ROBOTIQ2F_HARDWARE_INTERFACE_HPP