#include "robotiq_2f_interface/hardware_interface.hpp"

namespace robotiq_interface
{

const std::string robot_ip_parameter_name = "robot_ip";
const std::string robot_ip_default = "192.168.1.1";

const std::string robot_port_parameter_name = "robot_port";
const int robot_port_default = 63352;

const std::string use_mock_hardware_parameter_name = "use_fake_hardware";
const bool use_mock_hardware_default = true;

const std::string min_position_parameter_name = "min_position";
const double min_position_default = 0.01;

const std::string max_position_parameter_name = "max_position";
const double max_position_default = 0.0;

const std::string max_angle_parameter_name = "max_angle";
const double max_angle_default = 0.5;

const std::string min_speed_parameter_name = "min_speed";
const double min_speed_default = 0.0;

const std::string max_speed_parameter_name = "max_speed";
const double max_speed_default = 0.0;

const std::string min_force_parameter_name = "min_force";
const double min_force_default = 0.0;

const std::string max_force_parameter_name = "max_force";
const double max_force_default = 0.0;

RobotiqGripperHardwareInterface::RobotiqGripperHardwareInterface()
  : communication_thread_is_running_(false)
{
RCLCPP_DEBUG(logger_, "Constructor");
}

RobotiqGripperHardwareInterface::~RobotiqGripperHardwareInterface()
{
    if (communication_thread_is_running_.load())
    {
        communication_thread_is_running_.store(false);
        if (communication_thread_.joinable())
        {
            communication_thread_.join();
        }
    }
}

hardware_interface::CallbackReturn RobotiqGripperHardwareInterface::on_init(const hardware_interface::HardwareInfo& info)
{
    RCLCPP_DEBUG(logger_, "on_init");

    if (hardware_interface::SystemInterface::on_init(info) != CallbackReturn::SUCCESS)
    {
        return CallbackReturn::ERROR;
    }

    try
    {
        if (info.hardware_parameters.count(robot_ip_parameter_name) > 0) {
            robot_ip_ = info.hardware_parameters.at(robot_ip_parameter_name);
        } else {
            RCLCPP_DEBUG(logger_, "Falling back on default ip!");
            robot_ip_ = robot_ip_default;
        }
        RCLCPP_INFO(logger_, "Accessing parameter: Robot IP, Value: %s", robot_ip_.c_str());

        if (info.hardware_parameters.count(robot_port_parameter_name) > 0) {
            robot_port_ = std::stoi(info.hardware_parameters.at(robot_port_parameter_name));
        } else {
            robot_port_ = robot_port_default;
        }
        RCLCPP_DEBUG(logger_, "Accessing parameter: Robot Port, Value: %d", robot_port_);

        std::string param_value;
        if (info.hardware_parameters.count(use_mock_hardware_parameter_name) > 0) {
            param_value = info.hardware_parameters.at(use_mock_hardware_parameter_name);
            use_mock_hardware_ = (param_value == "true" || param_value == "True");
        } else {
            use_mock_hardware_ = use_mock_hardware_default;
        }
        RCLCPP_DEBUG(logger_, "Accessing parameter: use fake hardware , Value: %s", param_value.c_str());

        if (info.hardware_parameters.count(min_position_parameter_name) > 0) {
            min_position_ = std::stod(info.hardware_parameters.at(min_position_parameter_name));
        } else {
            min_position_ = min_position_default;
        }
        RCLCPP_DEBUG(logger_, "Accessing parameter: min position, Value: %f", min_position_);

        if (info.hardware_parameters.count(max_position_parameter_name) > 0) {
            max_position_ = std::stod(info.hardware_parameters.at(max_position_parameter_name));
        } else {
            max_position_ = max_position_default;
        }
        RCLCPP_DEBUG(logger_, "Accessing parameter: max position, Value: %f", max_position_);

        if (info.hardware_parameters.count(max_angle_parameter_name) > 0) {
            max_angle_ = std::stod(info.hardware_parameters.at(max_angle_parameter_name));
        } else {
            RCLCPP_WARN(logger_, "max angle not found falling back to default value!");
            max_angle_ = max_angle_default;
        }
        RCLCPP_DEBUG(logger_, "Accessing parameter: max angle , Value: %f", max_angle_);

        if (info.hardware_parameters.count(min_speed_parameter_name) > 0) {
            min_speed_ = std::stod(info.hardware_parameters.at(min_speed_parameter_name));
        } else {
            min_speed_ = min_speed_default;
        }
        RCLCPP_DEBUG(logger_, "Accessing parameter: min speed, Value: %f", min_speed_);

        if (info.hardware_parameters.count(max_speed_parameter_name) > 0) {
            max_speed_ = std::stod(info.hardware_parameters.at(max_speed_parameter_name));
        } else {
            max_speed_ = max_speed_default;
        }
        RCLCPP_DEBUG(logger_, "Accessing parameter: max speed, Value: %f", max_speed_);

                if (info.hardware_parameters.count(min_force_parameter_name) > 0) {
            min_force_ = std::stod(info.hardware_parameters.at(min_force_parameter_name));
        } else {
            min_force_ = robot_port_default;
        }
        RCLCPP_DEBUG(logger_, "Accessing parameter: min force, Value: %f", min_force_);

        if (info.hardware_parameters.count(max_force_parameter_name) > 0) {
            max_force_ = std::stod(info.hardware_parameters.at(max_force_parameter_name));
        } else {
            max_force_ = max_force_default;
        }
        RCLCPP_DEBUG(logger_, "Accessing parameter: max force, Value: %f", max_force_);
    }
    catch (const std::exception& e)
    {
        RCLCPP_FATAL(logger_, "Failed to load parameters: %s", e.what());
        return CallbackReturn::ERROR;
    }

    gripper_gap_ = std::numeric_limits<double>::quiet_NaN();
    gripper_angle_ = std::numeric_limits<double>::quiet_NaN();
    gripper_velocity_ = std::numeric_limits<double>::quiet_NaN();
    gripper_gap_command_ = std::numeric_limits<double>::quiet_NaN();
    reactivate_gripper_cmd_ = NO_NEW_CMD_;
    reactivate_gripper_async_cmd_.store(false);

    const hardware_interface::ComponentInfo& joint_gap = info_.joints[0];
    const hardware_interface::ComponentInfo& joint_angle = info_.joints[1];

    // There is one command interface: position.
    if (joint_gap.command_interfaces.size() != 1)
    {
        RCLCPP_FATAL(logger_, "Joint '%s' has %zu command interfaces found. 1 expected.", joint_gap.name.c_str(),
                     joint_gap.command_interfaces.size());
        return CallbackReturn::ERROR;
    }

    if (joint_gap.command_interfaces[0].name != hardware_interface::HW_IF_POSITION)
    {
        RCLCPP_FATAL(logger_, "Joint '%s' has %s command interfaces found. '%s' expected.", joint_gap.name.c_str(),
                     joint_gap.command_interfaces[0].name.c_str(), hardware_interface::HW_IF_POSITION);
        return CallbackReturn::ERROR;
    }

    // There are two state interfaces for the gripper gap: position and velocity.
    if (joint_gap.state_interfaces.size() != 2)
    {
        RCLCPP_FATAL(logger_, "Joint '%s' has %zu state interface. 2 expected.", joint_gap.name.c_str(),
                     joint_gap.state_interfaces.size());
        return CallbackReturn::ERROR;
    }

    for (int i = 0; i < 2; ++i)
    {
        if (!(joint_gap.state_interfaces[i].name == hardware_interface::HW_IF_POSITION ||
              joint_gap.state_interfaces[i].name == hardware_interface::HW_IF_VELOCITY))
        {
            RCLCPP_FATAL(logger_, "Joint '%s' has %s state interface. Expected %s or %s.", joint_gap.name.c_str(),
                       joint_gap.state_interfaces[i].name.c_str(), hardware_interface::HW_IF_POSITION,
                       hardware_interface::HW_IF_VELOCITY);
            return CallbackReturn::ERROR;
        }
    }

        // There are two state interfaces for the gripper angle: position.
    if (joint_angle.state_interfaces.size() != 1)
    {
        RCLCPP_FATAL(logger_, "Joint '%s' has %zu state interface. 1 expected.", joint_angle.name.c_str(),
                     joint_angle.state_interfaces.size());
        return CallbackReturn::ERROR;
    }

    if (!(joint_angle.state_interfaces[0].name == hardware_interface::HW_IF_POSITION))
    {
        RCLCPP_FATAL(logger_, "Joint '%s' has %s state interface. Expected %s or %s.", joint_angle.name.c_str(),
                   joint_angle.state_interfaces[0].name.c_str(), hardware_interface::HW_IF_POSITION,
                   hardware_interface::HW_IF_VELOCITY);
        return CallbackReturn::ERROR;
    }



    return CallbackReturn::SUCCESS;
}

hardware_interface::CallbackReturn RobotiqGripperHardwareInterface::on_configure(const rclcpp_lifecycle::State& previous_state)
{
    RCLCPP_DEBUG(logger_, "on_configure");
    try
    {
        if (hardware_interface::SystemInterface::on_configure(previous_state) != CallbackReturn::SUCCESS)
        {
            return CallbackReturn::ERROR;
        }

        configureAdapter(use_mock_hardware_);
    }
    catch (const std::exception& e)
    {
        RCLCPP_ERROR(logger_, "Cannot configure the Robotiq gripper: %s", e.what());
        return CallbackReturn::ERROR;
    }

    return CallbackReturn::SUCCESS;
}

hardware_interface::CallbackReturn RobotiqGripperHardwareInterface::on_activate(const rclcpp_lifecycle::State& /*previous_state*/)
{
    RCLCPP_DEBUG(logger_, "on_activate");

    // set some default values for joints
    if (std::isnan(gripper_gap_))
    {
        gripper_gap_ = 0.0;
        gripper_angle_ = 0.0;
        gripper_velocity_ = 0.0;
        gripper_gap_command_ = 0.0;
    }

    // Activate the gripper.
    try
    {
        RCLCPP_INFO(logger_, "Starting Socket connection!");
        if (!socket_adapter_->connecting(robot_ip_, robot_port_))
        {
            RCLCPP_ERROR(logger_, "Cannot connect to the Robotiq gripper at %s:%d", robot_ip_.c_str(), robot_port_);
            return CallbackReturn::ERROR;
        }
        socket_adapter_->deactivate();
        socket_adapter_->activate();

        communication_thread_is_running_.store(true);
        communication_thread_ = std::thread([this] { this->background_task(); });
        RCLCPP_INFO(logger_, "Background task thread started.");
    }
    catch (const std::exception& e)
    {
        RCLCPP_FATAL(logger_, "Failed to communicate with the Robotiq gripper: %s", e.what());
        return CallbackReturn::ERROR;
    }

    if (!socket_adapter_->is_active()) {
        RCLCPP_ERROR(logger_, "Attempted to activate while gripper is not active!");
        return CallbackReturn::ERROR;
    }
    else{
        RCLCPP_INFO(logger_, "Robotiq Gripper successfully activated!");
        return CallbackReturn::SUCCESS;
    }
}

hardware_interface::CallbackReturn RobotiqGripperHardwareInterface::on_deactivate(const rclcpp_lifecycle::State& /*previous_state*/)
{
    RCLCPP_DEBUG(logger_, "on_deactivate");

    communication_thread_is_running_.store(false);
    communication_thread_.join();
    if (communication_thread_.joinable())
    {
        communication_thread_.join();
    }

    try
    {
        socket_adapter_->deactivate();
    }
    catch (const std::exception& e)
    {
        RCLCPP_ERROR(logger_, "Failed to deactivate the Robotiq gripper: %s", e.what());
        return CallbackReturn::ERROR;
    }
    RCLCPP_INFO(logger_, "Robotiq Gripper successfully deactivated!");
    return CallbackReturn::SUCCESS;
}

std::vector<hardware_interface::StateInterface> RobotiqGripperHardwareInterface::export_state_interfaces()
{
    RCLCPP_DEBUG(logger_, "export_state_interfaces");

    std::vector<hardware_interface::StateInterface> state_interfaces;

    state_interfaces.emplace_back(
        hardware_interface::StateInterface(info_.joints[0].name, hardware_interface::HW_IF_POSITION, &gripper_gap_));
    state_interfaces.emplace_back(
        hardware_interface::StateInterface(info_.joints[0].name, hardware_interface::HW_IF_VELOCITY, &gripper_velocity_));
    state_interfaces.emplace_back(
        hardware_interface::StateInterface(info_.joints[1].name, hardware_interface::HW_IF_POSITION, &gripper_angle_));
    for (const auto& interface : state_interfaces) {
    RCLCPP_DEBUG(logger_, "Exporting state interface for joint: %s type: %s",
                interface.get_name().c_str(), interface.get_interface_name().c_str());
    }

    return state_interfaces;
}

std::vector<hardware_interface::CommandInterface> RobotiqGripperHardwareInterface::export_command_interfaces()
{
    RCLCPP_DEBUG(logger_, "export_command_interfaces");

    std::vector<hardware_interface::CommandInterface> command_interfaces;

    command_interfaces.emplace_back(
        hardware_interface::CommandInterface(
        info_.joints[0].name, hardware_interface::HW_IF_POSITION, &gripper_gap_command_
        )
    );

    // Handling potential uninitialized parameter 'gripper_speed_multiplier'
    double gripper_speed_multiplier = 1.0;
    if (info_.hardware_parameters.find("gripper_speed_multiplier") != info_.hardware_parameters.end()) {
        gripper_speed_multiplier = std::stod(info_.hardware_parameters["gripper_speed_multiplier"]);
    }
    gripper_speed_ = gripper_speed_multiplier;

    // Handling potential uninitialized parameter 'gripper_force_multiplier'
    double gripper_force_multiplier = 1.0;
    if (info_.hardware_parameters.find("gripper_force_multiplier") != info_.hardware_parameters.end()) {
        gripper_force_multiplier = std::stod(info_.hardware_parameters["gripper_force_multiplier"]);
    }
    gripper_force_ = gripper_force_multiplier;

    command_interfaces.emplace_back(
        hardware_interface::CommandInterface(info_.joints[0].name, "max_velocity", &gripper_speed_));
    command_interfaces.emplace_back(
        hardware_interface::CommandInterface(info_.joints[0].name, "max_effort", &gripper_force_));
    command_interfaces.emplace_back(
        hardware_interface::CommandInterface("reactivate_gripper", "reactivate_gripper_cmd", &reactivate_gripper_cmd_));
    command_interfaces.emplace_back(
        hardware_interface::CommandInterface("reactivate_gripper", "reactivate_gripper_response", &reactivate_gripper_response_));

    for (const auto& interface : command_interfaces) {
    RCLCPP_DEBUG(logger_, "Exporting command interface for joint: %s type: %s",
                interface.get_name().c_str(), interface.get_interface_name().c_str());
}

    return command_interfaces;
}

hardware_interface::return_type RobotiqGripperHardwareInterface::read(const rclcpp::Time& /*time*/, const rclcpp::Duration& /*period*/)
{
    RCLCPP_DEBUG(logger_, "Reading state from the gripper");

    // Ensure that the node uses the appropriate clock source

    current_time_ = clock_->now();

    try {
        int raw_position = socket_adapter_->position();
        double new_gap = convertRawToGap(raw_position);
        double new_angle = convertRawToAngle(raw_position);

        if (!std::isnan(gripper_gap_)) {
            double time_difference = (current_time_ - last_update_time_).seconds();
            if (time_difference > 0) {
                gripper_velocity_ = (new_gap - gripper_gap_) / time_difference;
            }
        }

        gripper_gap_ = new_gap;
        gripper_angle_ = new_angle;
        last_update_time_ = current_time_;

    } catch (const std::exception& e) {
        RCLCPP_ERROR(logger_, "Failed to read from Robotiq gripper: %s", e.what());
        return hardware_interface::return_type::ERROR;
    }

    // Process asynchronous reactivation response if available
    if (!std::isnan(reactivate_gripper_cmd_))
    {
        RCLCPP_INFO(logger_, "Sending gripper reactivation request.");
        reactivate_gripper_async_cmd_.store(true);
        reactivate_gripper_cmd_ = NO_NEW_CMD_;
    }

    if (reactivate_gripper_async_response_.load().has_value())
    {
        reactivate_gripper_response_ = reactivate_gripper_async_response_.load().value();
        reactivate_gripper_async_response_.store(std::nullopt);
    }

    return hardware_interface::return_type::OK;
}

hardware_interface::return_type RobotiqGripperHardwareInterface::write(const rclcpp::Time& /*time*/, const rclcpp::Duration& /*period*/)
{
    try {
        // Use the conversion methods to scale the command values to the range expected by the hardware
        int scaled_gap = convertGapToRaw(gripper_gap_command_);
        int scaled_speed = convertSpeedToRaw(gripper_speed_);
        int scaled_force = convertForceToRaw(gripper_force_);

        // Store the scaled values into atomic variables
        write_command_.store(scaled_gap);
        write_speed_.store(scaled_speed);
        write_force_.store(scaled_force);

    } catch (const std::exception& e) {
        RCLCPP_ERROR(logger_, "Failed to write to Robotiq gripper: %s", e.what());
        return hardware_interface::return_type::ERROR;
    }

    return hardware_interface::return_type::OK;
}

void RobotiqGripperHardwareInterface::background_task()
{
    while (communication_thread_is_running_.load())
    {
        std::lock_guard<std::mutex> guard(resource_mutex_);
        if (reactivate_gripper_async_cmd_.load())
        {
            try
            {
                // Execute the reactivation sequence
                RCLCPP_INFO(logger_, "Reactivating gripper.");
                socket_adapter_->deactivate();
                socket_adapter_->activate(false);
                reactivate_gripper_async_cmd_.store(false);  // Reset the command flag
                reactivate_gripper_async_response_.store(true);  // Store the success of the operation
            }
            catch (const std::exception& e)
            {
                RCLCPP_ERROR(logger_, "Failed to reactivate gripper: %s", e.what());
                reactivate_gripper_async_response_.store(false);  // Store the failure of the operation
            }
        }

        // Additional periodic operations
        performRegularOperations();
        std::this_thread::sleep_for(gripperCommsLoopPeriod);  // Sleep for the duration of the communication loop period
    }
}

void RobotiqGripperHardwareInterface::performRegularOperations()
{
    try
    {
        int scaled_gap = write_command_.load();
        int scaled_speed = write_speed_.load();
        int scaled_force = write_force_.load();

        if (scaled_gap != last_cmd_gap || scaled_speed != last_cmd_speed || scaled_force != last_cmd_force){
            RCLCPP_DEBUG(logger_, "New move cmd: POS: %d, SPE: %d, FOR: %d", scaled_gap, scaled_speed , scaled_force);
            auto result = socket_adapter_->move(scaled_gap, scaled_speed, scaled_force);
            if (!std::get<0>(result)) {
                throw std::runtime_error("Failed to send move command to Robotiq gripper.");
            }
            last_cmd_gap = scaled_gap;
            last_cmd_speed = scaled_speed;
            last_cmd_force = scaled_force;
        }
        int raw_gap = socket_adapter_->position();
        gripper_current_state_.store(convertRawToGap(raw_gap));
    }
    catch (const std::exception& e)
    {
        RCLCPP_ERROR(logger_, "Regular operation error: %s", e.what());
    }
}


// Mock Hardware
void RobotiqGripperHardwareInterface::configureAdapter(bool useMock) {
    if (useMock) {
        RCLCPP_INFO(logger_, "Using Mock Robotiq Adapter");
        socket_adapter_ = std::make_unique<MockRobotiq2fSocketAdapter>();
    } else {
        RCLCPP_INFO(logger_, "Using Real Robotiq Adapter");
        socket_adapter_ = std::make_unique<Robotiq2fSocketAdapter>();
    }
}


// Conversion methods
double RobotiqGripperHardwareInterface::convertRawToGap(int raw_position) {
    if (std::isnan(min_position_) || std::isnan(max_position_) || min_position_ >= max_position_) {
        throw std::runtime_error("Invalid gripper position limits.");
    }
    double scaled_position = min_position_ + ((230 - static_cast<double>(raw_position)) / 227.0) * (max_position_ - min_position_);
    return std::max(min_position_, std::min(scaled_position, max_position_));
}

double RobotiqGripperHardwareInterface::convertRawToAngle(int raw_position) {
    if (std::isnan(max_angle_) || 0.0 >= max_angle_) {
        throw std::runtime_error("Invalid gripper angle limits: max_angle is not configured correctly.");
    }
    double scaled_angle = (static_cast<double>(raw_position) / 255.0) * max_angle_;
    return std::max(0.0, std::min(scaled_angle, max_angle_));
}

unsigned int RobotiqGripperHardwareInterface::convertGapToRaw(double gap) {
    if (std::isnan(min_position_) || std::isnan(max_position_) || min_position_ >= max_position_) {
        throw std::runtime_error("Invalid gripper position limits.");
    }
    unsigned int raw = static_cast<unsigned int>(230 - ((gap - min_position_) / (max_position_ - min_position_)) * 227);
    return std::clamp(raw, 3u, 230u); // Ensuring the output is within the valid range
}

unsigned int RobotiqGripperHardwareInterface::convertSpeedToRaw(double speed) {
    if (std::isnan(min_speed_) || std::isnan(max_speed_) || min_speed_ >= max_speed_) {
        throw std::runtime_error("Invalid gripper speed limits: min_speed_ or max_speed_ are not configured correctly.");
    }
    double scaled = (speed - min_speed_) / (max_speed_ - min_speed_) * 255.0;
    return static_cast<unsigned int>(std::clamp(scaled, 0.0, 255.0));
}

unsigned int RobotiqGripperHardwareInterface::convertForceToRaw(double force) {
    if (std::isnan(min_force_) || std::isnan(max_force_) || min_force_ >= max_force_) {
        throw std::runtime_error("Invalid gripper force limits: min_force_ or max_force_ are not configured correctly.");
    }
    double scaled = (force - min_force_) / (max_force_ - min_force_) * 255.0;
    return static_cast<unsigned int>(std::clamp(scaled, 0.0, 255.0));
}

}  // namespace robotiq_interface

#include "pluginlib/class_list_macros.hpp"

PLUGINLIB_EXPORT_CLASS(robotiq_interface::RobotiqGripperHardwareInterface, hardware_interface::SystemInterface)
